#include <fftw3.h>
#include <stdlib.h>

#define INVERSE +1
#define FORWARD -1
/* Using LAPACK notation for double/single precision complex */
void zfft1d(fftw_complex *zin, fftw_complex *zout, 
	    int len_xform, int len_z, int direction, int shift);
void cfft1d(fftwf_complex *zin, fftwf_complex *zout, 
	    int len_xform, int len_z, int direction, int shift);
void zfft2d(fftw_complex *zin, fftw_complex *zout, int xdim, int ydim,
	    int len_z, int direction, int shift);
void cfft2d(fftwf_complex *zin, fftwf_complex *zout, int xdim, int ydim,
	    int len_z, int direction, int shift);

/**************************************************************************
* (z/c)fft1d                                                              *
*                                                                         *
* Repeatedly takes a 1D FFT of length len_xform by advancing the zin and  *
* zout pointers len_z/len_xform times. "direction" indicates whether the  *
* transform is a forward of reverse FFT. Quadrant shifting a la fftshift  *
* in Matlab is done by modulation in both domains. Always normalize ifft. *
**************************************************************************/

void zfft1d(fftw_complex *zin, fftw_complex *zout, 
	    int len_xform, int len_z, int direction, int shift)
{
  fftw_plan FT1D;
  fftw_complex *zptr1=NULL, *zptr2=NULL;
  double tog = 1.0, alpha;
  int k, nxforms = len_z/len_xform;

  FT1D = fftw_plan_many_dft(1, &len_xform, nxforms, zin, NULL, 1, len_xform, 
			    zout, NULL, 1, len_xform, direction,
			    FFTW_ESTIMATE | FFTW_PRESERVE_INPUT);
  // if shifting, modulate the input
  if(shift) {
    zptr1 = zin;
    for(k=0; k<len_z; k++) {
      *((double *) (zptr1)) *= tog;
      *((double *) (zptr1++) + 1) *= tog;
      tog *= -1.0;
    }
  }
  fftw_execute(FT1D);
  tog = 1.0;
  // demodulate output (and restore input if it's a separate array)
  zptr2 = zout;
  if(shift) {
    zptr1 = (zin != zout) ? zin : NULL;
    for(k=0; k<len_z; k++) {
      if(zptr1) {
	*((double *) (zptr1)) *= tog;
	*((double *) (zptr1++) + 1) *= tog;
      }
      alpha = (direction==INVERSE) ? tog/(double) len_xform : tog;
      *((double *) (zptr2)) *= alpha;
      *((double *) (zptr2++) + 1) *= alpha;
      tog *= -1.0;
    }
  } else {
    if(direction == INVERSE) {
      alpha = 1.0 / (double) len_xform;
      for(k=0; k<len_z; k++) {
	*((double *) (zptr2)) *= alpha;
	*((double *) (zptr2++) + 1) *= alpha;
      }
    }
  }
  fftw_destroy_plan(FT1D);
  //fftw_cleanup();
}

void cfft1d(fftwf_complex *zin, fftwf_complex *zout, 
	    int len_xform, int len_z, int direction, int shift)
{
  fftwf_plan FT1D;
  fftwf_complex *cptr1=NULL, *cptr2=NULL;
  float tog = 1.0, alpha;
  int k, nxforms = len_z/len_xform;
  FT1D = fftwf_plan_many_dft(1, &len_xform, nxforms, zin, NULL, 1, len_xform, 
			     zout, NULL, 1, len_xform, direction,
			     FFTW_ESTIMATE | FFTW_PRESERVE_INPUT);

  // modulate the input
  if(shift) {
    cptr1 = zin;
    for(k=0; k<len_z; k++) {
      *((float *) (cptr1)) *= tog;
      *((float *) (cptr1++) + 1) *= tog;
      tog *= -1.0;
    }
  }
  fftwf_execute(FT1D);
  tog = 1.0;
  // demodulate output (and restore input if it's a separate array)
  cptr2 = zout;
  if(shift) {
    cptr1 = (zin != zout) ? zin : NULL;
    for(k=0; k<len_z; k++) {
      if(cptr1) {
	*((float *) (cptr1)) *= tog;
	*((float *) (cptr1++) + 1) *= tog;	
      }
      alpha = (direction==INVERSE) ? tog/(float) len_xform : tog;
      *((float *) (cptr2)) *= alpha;
      *((float *) (cptr2++) + 1) *= alpha;
      tog *= -1.0;
    }
  } else {
    if(direction == INVERSE) {
      alpha = 1.0 / (float) len_xform;
      for(k=0; k<len_z; k++) {
	*((float *) (cptr2)) *= alpha;
	*((float *) (cptr2++) + 1) *= alpha;
      }
    }
  }
  fftwf_destroy_plan(FT1D);
  //fftwf_cleanup();
}


/**************************************************************************
* (z/c)fft2d                                                              *
*                                                                         *
* Repeatedly take a 2D FFT of length xdim*ydim across the input array.    *
* "direction" incidates the sign on the complex exponential's argument.   *
* Quadrant "shifting" a la fftshift in Matlab is done by modulation in    *
* both domains. Always normalize the output of the inverse FFT.           *
**************************************************************************/
void zfft2d(fftw_complex *zin, fftw_complex *zout, int xdim, int ydim,
	    int len_z, int direction, int shift)
{
  double tog = 1.0, alpha;
  int k, size_xform = xdim*ydim, nxforms = len_z/(xdim*ydim);
  int dims[2] = {ydim, xdim};
  fftw_plan FT2D;
  fftw_complex *zptr1, *zptr2;
  
  FT2D = fftw_plan_many_dft(2, dims, nxforms, 
			    zin, NULL, 1, size_xform,
			    zout, NULL, 1, size_xform,
			    direction, FFTW_ESTIMATE | FFTW_PRESERVE_INPUT);
  if(shift) {
    zptr1 = zin;
    for(k=0; k<len_z; k++) {
      *((double *) (zptr1)) *= tog;
      *((double *) (zptr1++) + 1) *= tog;
      if( (k+1)%xdim ) {
	tog *= -1.0;
      }
    }
  }
  fftw_execute(FT2D);
  tog = 1.0;
  zptr2 = zout;
  if(shift) {
    zptr1 = (zin != zout) ? zin : NULL;
    for(k=0; k<len_z; k++) {
      if(zptr1) {
	*((double *) (zptr1)) *= tog;
	*((double *) (zptr1++) + 1) *= tog;
      }
      alpha = (direction==INVERSE) ? tog/(double) size_xform : tog;
      *((double *) (zptr2)) *= alpha;
      *((double *) (zptr2++) + 1) *= alpha;
      if( (k+1)%xdim ) {
	tog *= -1.0;
      }
    }
  } else {
    if(direction==INVERSE) {
      alpha = 1.0 / (double) size_xform;
      for(k=0; k<len_z; k++) {
	*((double *) (zptr2)) *= alpha;
	*((double *) (zptr2++) + 1) *= alpha;
      }
    }
  }
    
  fftw_destroy_plan(FT2D);
  //fftw_cleanup();
  return;
}	     

void cfft2d(fftwf_complex *zin, fftwf_complex *zout, int xdim, int ydim,
	    int len_z, int direction, int shift)
{
  float tog = 1.0, alpha;
  int k, size_xform = xdim*ydim, nxforms = len_z/(xdim*ydim);
  int dims[2] = {ydim, xdim};
  fftwf_plan FT2D;
  fftwf_complex *cptr1, *cptr2;

  FT2D = fftwf_plan_many_dft(2, dims, nxforms, 
			     zin, NULL, 1, size_xform,
			     zout, NULL, 1, size_xform,
			     direction, FFTW_ESTIMATE | FFTW_PRESERVE_INPUT);
  if(shift) {
    cptr1 = zin;
    for(k=0; k<len_z; k++) {
      *((float *) (cptr1)) *= tog;
      *((float *) (cptr1++) + 1) *= tog;
      if( (k+1)%xdim ) {
	tog *= -1.0;
      }
    }
  }
  fftwf_execute(FT2D);
  tog = 1.0;
  cptr2 = zout;
  if(shift) {
    cptr1 = (zin != zout) ? zin : NULL;
    for(k=0; k<len_z; k++) {
      if(cptr1) {
	*((float *) (cptr1)) *= tog;
	*((float *) (cptr1++) + 1) *= tog;
      }
      alpha = (direction==INVERSE) ? tog/(float) size_xform : tog;
      *((float *) (cptr2)) *= alpha;
      *((float *) (cptr2++) + 1) *= alpha;
      if( (k+1)%xdim ) {
	tog *= -1.0;
      }
    }
  } else {
    if(direction==INVERSE) {
      alpha = 1.0 / (float) size_xform;
      for(k=0; k<len_z; k++) {
	*((float *) (cptr2)) *= alpha;
	*((float *) (cptr2++) + 1) *= alpha;
      }
    }
  }

  fftwf_destroy_plan(FT2D);
  //fftwf_cleanup();
  return;
}	     


/* static struct PyMethodDef fftmod_module_methods[] = { */
/*   {"fft1d", (PyCFunction)fftmod_fft1d, 1, doc_fft1d}, */
/*   {"fft2d", (PyCFunction)fftmod_fft2d, 1, doc_fft2d}, */
/*   {NULL, NULL, 0} */
/* }; */

/* DL_EXPORT(void) init_fftmod(void) { */
/*   PyObject *m; */
/*   m = Py_InitModule3("_fftmod", fftmod_module_methods, "C library for complex FFT"); */
/*   import_array(); */
/* } */
